/**
 * This file Copyright (c) 2010-2018 Magnolia International
 * Ltd.  (http://www.magnolia-cms.com). All rights reserved.
 *
 *
 * This file is dual-licensed under both the Magnolia
 * Network Agreement and the GNU General Public License.
 * You may elect to use one or the other of these licenses.
 *
 * This file is distributed in the hope that it will be
 * useful, but AS-IS and WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE, TITLE, or NONINFRINGEMENT.
 * Redistribution, except as permitted by whichever of the GPL
 * or MNA you select, is prohibited.
 *
 * 1. For the GPL license (GPL), you can redistribute and/or
 * modify this file under the terms of the GNU General
 * Public License, Version 3, as published by the Free Software
 * Foundation.  You should have received a copy of the GNU
 * General Public License, Version 3 along with this program;
 * if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * 2. For the Magnolia Network Agreement (MNA), this file
 * and the accompanying materials are made available under the
 * terms of the MNA which accompanies this distribution, and
 * is available at http://www.magnolia-cms.com/mna.html
 *
 * Any modifications to this file must keep this entire header
 * intact.
 *
 */
package info.magnolia.ui.vaadin.gwt.client.connector;

import info.magnolia.ui.vaadin.editor.PageEditor;
import info.magnolia.ui.vaadin.gwt.client.css.PageEditorCssProvider;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.CmsNode;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.Comment;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.MgnlComponent;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.MgnlElement;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.processor.AbstractMgnlElementProcessor;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.processor.CommentProcessor;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.processor.ElementProcessor;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.processor.MgnlElementProcessorFactory;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.processor.ProcessException;
import info.magnolia.ui.vaadin.gwt.client.editor.event.ComponentActionEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.event.ComponentStartMoveEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.event.ComponentStopMoveEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.event.EditAreaEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.event.EditComponentEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.event.FrameNavigationEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.event.NewAreaEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.event.NewComponentEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.event.SelectElementEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.event.SortComponentEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.jsni.event.FrameLoadedEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.model.Model;
import info.magnolia.ui.vaadin.gwt.client.editor.model.ModelImpl;
import info.magnolia.ui.vaadin.gwt.client.editor.model.focus.FocusModel;
import info.magnolia.ui.vaadin.gwt.client.editor.model.focus.FocusModelImpl;
import info.magnolia.ui.vaadin.gwt.client.rpc.PageEditorClientRpc;
import info.magnolia.ui.vaadin.gwt.client.rpc.PageEditorServerRpc;
import info.magnolia.ui.vaadin.gwt.client.shared.AbstractElement;
import info.magnolia.ui.vaadin.gwt.client.shared.AreaElement;
import info.magnolia.ui.vaadin.gwt.client.shared.ComponentElement;
import info.magnolia.ui.vaadin.gwt.client.shared.PageElement;
import info.magnolia.ui.vaadin.gwt.client.widget.PageEditorView;
import info.magnolia.ui.vaadin.gwt.client.widget.PageEditorViewImpl;
import info.magnolia.ui.vaadin.gwt.client.widget.dnd.MoveWidget;

import java.util.List;
import java.util.logging.Logger;

import com.google.gwt.core.client.GWT;
import com.google.gwt.core.client.JavaScriptException;
import com.google.gwt.dom.client.Document;
import com.google.gwt.dom.client.Element;
import com.google.gwt.dom.client.HeadElement;
import com.google.gwt.dom.client.LinkElement;
import com.google.gwt.dom.client.Node;
import com.google.gwt.event.shared.EventBus;
import com.google.gwt.event.shared.SimpleEventBus;
import com.google.gwt.regexp.shared.MatchResult;
import com.google.gwt.regexp.shared.RegExp;
import com.google.gwt.user.client.ui.Widget;
import com.vaadin.client.BrowserInfo;
import com.vaadin.client.communication.RpcProxy;
import com.vaadin.client.communication.StateChangeEvent;
import com.vaadin.client.ui.AbstractComponentConnector;
import com.vaadin.shared.ui.Connect;

/**
 * Client side connector which connects to {@link PageEditor}.
 */
@Connect(PageEditor.class)
public class PageEditorConnector extends AbstractComponentConnector implements PageEditorView.Listener {

    private Logger log = Logger.getLogger(getClass().getName());

    private static final String IFRAME_PRELOADER_STYLE = "iframe-preloader";

    private final PageEditorServerRpc rpc = RpcProxy.create(PageEditorServerRpc.class, this);

    private final EventBus eventBus = new SimpleEventBus();

    private PageEditorView view;
    private MoveWidget moveWidget;

    private Model model;

    private FocusModel focusModel;
    private ElementProcessor elementProcessor;
    private CommentProcessor commentProcessor;

    // This indicates whether the page-editor is about to open a new page, by following a link inside the iframe.
    // As the selected element read from the state, won't be up-to-date until the server-side has been notified about
    // the page-change and pushed by the state to the client, we set the selected element to null when the page is loaded
    // to make the client fall back to the page-element.
    // See FrameNavigationEventHandler#onFrameUrlChanged() and FrameLoadedEvent.Handler#handle()
    private boolean pageChange = false;

    @Override
    protected void init() {
        super.init();
        this.model = new ModelImpl();
        this.focusModel = new FocusModelImpl(eventBus, model);
        this.elementProcessor = new ElementProcessor(eventBus, model);
        this.commentProcessor = new CommentProcessor();

        addStateChangeHandler("url", (StateChangeEvent.StateChangeHandler) stateChangeEvent -> {
            view.setUrl(getState().getUrl());
            if (getState().isPreview()) {
                view.getFrame().addStyleName(IFRAME_PRELOADER_STYLE);
            }
        });
        registerRpc(PageEditorClientRpc.class, new PageEditorClientRpc() {

            @Override
            public void refresh() {
                view.reload();
            }

            @Override
            public void startMoveComponent() {
                MgnlElement selectedElement = model.getMgnlElement(getState().getSelectedElement());
                if (selectedElement != null && selectedElement.isComponent()) {
                    MgnlComponent component = (MgnlComponent) selectedElement;
                    component.doStartMove(false);
                    model.setMoving(true);
                    if (!BrowserInfo.get().isTouchDevice()) {
                        PageEditorConnector.this.moveWidget = component.getMoveWidget();
                        moveWidget.attach(view.getFrame());
                    }
                }
            }

            @Override
            public void cancelMoveComponent() {
                eventBus.fireEvent(new ComponentStopMoveEvent(null, true));
            }
        });

        /**
         * The {@link FrameLoadedEvent} is thrown in {@link info.magnolia.ui.vaadin.gwt.client.editor.jsni.AbstractFrameEventHandler#onFrameReady()}
         * and indicates that the iFrame has been loaded and the document inside it is ready to be parsed.
         */
        eventBus.addHandler(FrameLoadedEvent.TYPE, event -> {
            model.reset();
            focusModel.resetFocus();
            Document document = null;
            try {
                document = event.getFrame().getContentDocument();
            } catch (JavaScriptException e) {
                GWT.log("Error getting content document from iframe: " + e.getMessage());
            }
            if (document == null) {
                rpc.selectExternalPage();
            } else {
                process(document);

                view.initKeyEventListeners();

                if (!getState().isPreview()) {
                    view.initDomEventListeners();
                    focusModel.init();
                }
                AbstractElement elementToSelect = pageChange ? null : getState().getSelectedElement();
                focusModel.selectElement(elementToSelect);
            }
            PageEditorConnector.this.pageChange = false;

            if (event.getFrame() != null) {
                rpc.onNavigation(removePageEditorParameters(event.getFrame().getUrl()));
            }
        });

        eventBus.addHandler(FrameNavigationEvent.TYPE, frameUrlChangedEvent -> {
            String path = frameUrlChangedEvent.getPath();
            RegExp regExp = RegExp.compile("((?:([^:/?#]+):)?(?://([^/?#]*))?([^?#]*))(?:(\\?[^#]*))?(?:(#.*))?");
            MatchResult matchResult = regExp.exec(path);
            if (matchResult != null) {
                path = matchResult.getGroup(1);

                final String platformId = getState().getPlatformType().getId();
                final boolean isPreview = getState().isPreview();
                String query = matchResult.getGroup(5);
                query = query != null ? removePageEditorParameters(query) : query;
                query = query != null && query != "" ? query + "&" : "?";
                query += "mgnlChannel=" + platformId;
                query += "&mgnlPreview=" + isPreview;
                path += query;

                String fragment = matchResult.getGroup(6);
                if (fragment != null) {
                    path += fragment;
                }
            }

            PageEditorConnector.this.pageChange = true;
            view.setUrl(path);
        });

        eventBus.addHandler(SelectElementEvent.TYPE, selectElementEvent -> {
            AbstractElement selectedElement = selectElementEvent.getElement();
            if (selectedElement instanceof PageElement) {
                rpc.selectPage((PageElement) selectedElement);
            } else if (selectedElement instanceof AreaElement) {
                rpc.selectArea((AreaElement) selectedElement);
            } else if (selectedElement instanceof ComponentElement) {
                rpc.selectComponent((ComponentElement) selectedElement);
            }
            view.resetScrollTop();
        });

        eventBus.addHandler(NewAreaEvent.TYPE, newAreaEvent -> {
            if (!getConnection().getMessageSender().hasActiveRequest()) {
                rpc.newArea(newAreaEvent.getAreaElement());
            }
        });

        eventBus.addHandler(NewComponentEvent.TYPE, newComponentEvent -> {
            if (!getConnection().getMessageSender().hasActiveRequest()) {
                rpc.newComponent(newComponentEvent.getParentAreaElement());
            }
        });

        eventBus.addHandler(EditAreaEvent.TYPE, editAreaEvent -> {
            if (!getConnection().getMessageSender().hasActiveRequest()) {
                rpc.editArea(editAreaEvent.getAreaElement());
            }
        });

        eventBus.addHandler(EditComponentEvent.TYPE, editComponentEvent -> {
            if (!getConnection().getMessageSender().hasActiveRequest()) {
                rpc.editComponent(editComponentEvent.getComponentElement());
            }
        });

        eventBus.addHandler(SortComponentEvent.TYPE, sortComponentEvent -> rpc.sortComponent(sortComponentEvent.getAreaElement()));

        eventBus.addHandler(ComponentStartMoveEvent.TYPE, componentStartMoveEvent -> rpc.startMoveComponent());

        eventBus.addHandler(ComponentStopMoveEvent.TYPE, componentStopMoveEvent -> {
            if (!componentStopMoveEvent.isServerSide()) {
                rpc.stopMoveComponent();
            }
            if (moveWidget != null && moveWidget.isAttached()) {
                moveWidget.detach();
            }
            model.setMoving(false);
        });

        eventBus.addHandler(ComponentActionEvent.TYPE, actionEvent -> rpc.onClientAction(actionEvent.getElement(), actionEvent.getActionName(), actionEvent.getParameters()));
    }

    @Override
    protected Widget createWidget() {
        this.view = new PageEditorViewImpl(eventBus);
        this.view.setListener(this);
        return view.asWidget();
    }

    @Override
    public PageEditorState getState() {
        return (PageEditorState) super.getState();
    }

    @Override
    public void selectElement(Element element) {
        focusModel.selectElement(element);
    }

    private void process(final Document document) {
        try {
            injectEditorStyles(document);
            long startTime = System.currentTimeMillis();
            processDocument(document, null);
            processMgnlElements();
            GWT.log("Time spent to process cms comments: " + (System.currentTimeMillis() - startTime) + "ms");
        } catch (ProcessException e) {
            rpc.onError(e.getErrorType(), e.getTagName());
            GWT.log("Error while processing comment: " + e.getTagName() + " due to " + e.getErrorType());
            consoleLog("Error while processing comment: " + e.getTagName() + " due to " + e.getErrorType()); // log also into browser console
        }
    }

    private void injectEditorStyles(final Document document) {
        HeadElement head = HeadElement.as(document.getElementsByTagName("head").getItem(0));
        for (String uri: PageEditorCssProvider.INSTANCE.getCssLinks()) {
            LinkElement cssLink = createLinkElement(document, getState().getContextPath() + uri);
            head.insertFirst(cssLink);
        }
    }

    private LinkElement createLinkElement(final Document document, final String href) {
        LinkElement cssLink = document.createLinkElement();
        cssLink.setType("text/css");
        cssLink.setRel("stylesheet");
        cssLink.setHref(href);
        return cssLink;
    }

    private void processDocument(Node node, MgnlElement mgnlElement) throws ProcessException {
        if (mgnlElement == null && model.getRootPage() != null) {
            mgnlElement = model.getRootPage();
        }
        for (int i = 0; i < node.getChildCount(); i++) {
            Node childNode = node.getChild(i);
            if (childNode.getNodeType() == Comment.COMMENT_NODE) {
                try {
                    mgnlElement = commentProcessor.process(model, eventBus, childNode, mgnlElement);
                } catch (ProcessException e) {
                    throw e;
                } catch (IllegalArgumentException e) {
                    GWT.log("Not CMSComment element, skipping: " + e.toString());
                } catch (Exception e) {
                    GWT.log("Caught undefined exception: " + e.toString());
                    consoleLog("Caught undefined exception: " + e.toString()); // log also into browser console
                }
            } else if (childNode.getNodeType() == Node.ELEMENT_NODE) {
                Element element = childNode.cast();
                elementProcessor.process(element, mgnlElement, getState().isPreview());
            }
            processDocument(childNode, mgnlElement);
        }
    }

    private void processMgnlElements() {
        CmsNode root = model.getRootPage();
        if (model.getRootPage() == null) {
            log.warning("Could not find any Magnolia cms:page tag, this might be a static page; not injecting page-editor bars.");
            return;
        }

        List<CmsNode> elements = root.getDescendants();
        elements.add(root);
        for (CmsNode element : elements) {
            try {
                AbstractMgnlElementProcessor processor = MgnlElementProcessorFactory.getProcessor(model, element.asMgnlElement(), getState().getI18nKeys());
                processor.process();
            } catch (IllegalArgumentException e) {
                GWT.log("MgnlFactory could not instantiate class. The element is neither an area nor component.");
            } catch (Exception e) {
                final String errorMessage = "Error when processing editor components for '" + element.asMgnlElement().getAttribute("path") +
                        "'. It's possible that the template script for this area or for some subcomponent is incorrect. Please check that all HTML tags are closed properly.\n"
                        + e.toString();
                GWT.log(errorMessage);
                consoleLog(errorMessage); // log also into browser console
            }
        }
    }

    void consoleLog(String message) {
        log.info("PageEditor: " + message);
    }

    /**
     * Clean up mgnlChannel parameter & mgnlPreview parameter, those param used for page editor only.
     * The algorithm is to firstly remove mgnlChannel and mgnlPreview without '?', then remove '?' if it stands alone or remove '&' if '?' combine with '&'.
     * <ul>
     * This is example:
     * <li> Input: http://localhost:8185/g.html?mgnlChannel=ee&a=a&mgnlPreview=pp
     * <li> Replace steps:
     * <ul>
     * <li> Step 1: http://localhost:8185/g.html?&a=a
     * <li> Step 2: http://localhost:8185/g.html?&a=a
     * <li> Step 3: http://localhost:8185/g.html?a=a
     * </ul>
     * </ul>
     */
    private String removePageEditorParameters(String url) {
        return url.replaceAll("(\\&{0,1})(mgnlChannel|mgnlPreview)\\=([^&]+)", "").replaceAll("(\\?$)", "").replaceAll("(\\?\\&)", "?");
    }
}
