/**
 * This file Copyright (c) 2012-2018 Magnolia International
 * Ltd.  (http://www.magnolia-cms.com). All rights reserved.
 *
 *
 * This file is dual-licensed under both the Magnolia
 * Network Agreement and the GNU General Public License.
 * You may elect to use one or the other of these licenses.
 *
 * This file is distributed in the hope that it will be
 * useful, but AS-IS and WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE, TITLE, or NONINFRINGEMENT.
 * Redistribution, except as permitted by whichever of the GPL
 * or MNA you select, is prohibited.
 *
 * 1. For the GPL license (GPL), you can redistribute and/or
 * modify this file under the terms of the GNU General
 * Public License, Version 3, as published by the Free Software
 * Foundation.  You should have received a copy of the GNU
 * General Public License, Version 3 along with this program;
 * if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * 2. For the Magnolia Network Agreement (MNA), this file
 * and the accompanying materials are made available under the
 * terms of the MNA which accompanies this distribution, and
 * is available at http://www.magnolia-cms.com/mna.html
 *
 * Any modifications to this file must keep this entire header
 * intact.
 *
 */
package info.magnolia.ui.vaadin.gwt.client.editor.model.focus;

import info.magnolia.ui.vaadin.gwt.client.editor.dom.MgnlArea;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.MgnlComponent;
import info.magnolia.ui.vaadin.gwt.client.editor.dom.MgnlElement;
import info.magnolia.ui.vaadin.gwt.client.editor.event.ComponentStopMoveEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.event.SelectElementEvent;
import info.magnolia.ui.vaadin.gwt.client.editor.jsni.scroll.ElementScrollPositionPreserver;
import info.magnolia.ui.vaadin.gwt.client.editor.model.Model;
import info.magnolia.ui.vaadin.gwt.client.shared.AbstractElement;

import com.google.gwt.dom.client.Element;
import com.google.web.bindery.event.shared.EventBus;

/**
 * Takes care of toggling the visibility of the control-bars in the page-editor based on selections of the elements.
 * This is a fairly optimized and complex implementation considering lots of special corner-cases and side-effects. Don't
 * apply changes here lightly. Seriously.
 * This class should remain state-less. Ideally it should only contain the logic for switching the visibility of
 * control-bars inside the iframe.
 */
public class FocusModelImpl implements FocusModel {

    private final Model model;
    private final EventBus eventBus;

    // These elements are only tracked for the toggling of visibility.
    // They should not survive a reload of the page and should not be exposed.
    private MgnlArea selectedAreaElement;
    private MgnlComponent selectedComponentElement;

    public FocusModelImpl(EventBus eventBus, Model model) {
        this.eventBus = eventBus;
        this.model = model;
    }

    @Override
    public void selectElement(Element element) {
        selectElement(model.getMgnlElement(element));
    }

    @Override
    public void selectElement(AbstractElement element) {
        selectElement(model.getMgnlElement(element));
    }

    /**
     * Selects the given {@link MgnlElement element} and keeps track of the scroll position before toggling the control-bars
     * for later re-adjustments.
     */
    private void selectElement(MgnlElement mgnlElement) {

        if (mgnlElement == null) {
            mgnlElement = model.getRootPage();
        }

        // cancel move if click was outside the moving components relatives
        if (model.isMoving()) {
            if (!mgnlElement.isRelated(selectedComponentElement)) {
                eventBus.fireEvent(new ComponentStopMoveEvent(null, false));
            }
            return; // just cancel the move, no selection
        }

        // reset the scroll position unless it's a page.
        ElementScrollPositionPreserver scrollPositionPreserver = null;
        if (!mgnlElement.isPage()) {
            scrollPositionPreserver = mgnlElement.preserve();
        }

        doSelectMgnlElement(mgnlElement);

        // restoring of the position has to happen after the selection, as that's when the page-height has been changed.
        if (scrollPositionPreserver != null) {
            scrollPositionPreserver.restorePosition();
        }

    }

    private void doSelectMgnlElement(MgnlElement mgnlElement) {
        MgnlArea area = null;
        MgnlComponent component = null;

        if (mgnlElement.isComponent()) {
            component = (MgnlComponent) mgnlElement;
            area = mgnlElement.getParentArea();
        } else if (mgnlElement.isArea()) {
            area = (MgnlArea) mgnlElement;
        }
        // first set the component, then set the area. the selected component is used for setting
        // the current area class.
        setComponentSelection(component);
        setAreaSelection(area);

        // if we're selecting the page, we want this light-greenish look on the root areas, if not, then not.
        togglePageFocus(mgnlElement.isPage());

        dispatchElementSelection(mgnlElement);
    }

    /**
     * Creates the initial selection of the page-editor when in editing-mode:
     * <pre>
     *     <ul>
     *         <li>Root-areas are visible</li>
     *         <li>Component-placeholders in the root-areas are visible.</li>
     *     </ul>
     * </pre>
     *
     * Opposed to {@link #togglePageFocus(boolean)} this method only needs to be called once, when the iframe is loaded.
     * The reason is that these settings will never be reset by any selections done afterwards.
     */
    @Override
    public void init() {
        for (MgnlArea root : model.getRootAreas()) {
            root.setVisible(true);

            if (root.getComponents().isEmpty()) {
                root.setPlaceHolderVisible(true);
            }
        }
    }

    /**
     * Resets the last selected elements. This needs to happen as lot's of the toggling-logic of the displayed bars and
     * their status is based on these. It has no visual effect. This method should ideally go, as the selected elements
     * should automatically be disposed, when the dom changes, e.g. when the iframe reloads.
     */
    @Override
    public void resetFocus() {
        this.selectedAreaElement = null;
        this.selectedComponentElement = null;
    }

    /**
     * Takes care of the selection of components. keeps track of last selected element and toggles
     * the focus. If a null-value is passed it will reset the currently selected component.
     *
     * @param component the MgnlElement component, can be null.
     */
    private void setComponentSelection(MgnlComponent component) {
        if (selectedComponentElement == component) {
            return;
        }
        if (selectedComponentElement != null) {
            selectedComponentElement.removeFocus();
        }
        if (component != null) {
            component.setFocus();
        }
        this.selectedComponentElement = component;
    }

    /**
     * This method takes care of selecting and deselecting areas.
     *
     * @param area selected area, can be null.
     */
    private void setAreaSelection(MgnlArea area) {

        if (selectedAreaElement != null) {

            selectedAreaElement.removeFocus();

            // always reset current area selection unless area and current area are related
            if (!selectedAreaElement.isRelated(area)) {

                toggleChildComponentVisibility(selectedAreaElement, false);
                toggleAreaVisibility(selectedAreaElement, false);
            }

            // hide child components if area is an ascendant of current area or selectedArea is not
            // a descendant
            else if (selectedAreaElement.getAscendants().contains(area) || (area != null && !area.getDescendants().contains(selectedAreaElement))) {
                toggleChildComponentVisibility(selectedAreaElement, false);
            }
        }

        // set focus on new selected area
        if (area != null) {
            toggleAreaVisibility(area, true);
            toggleChildComponentVisibility(area, true);

            area.setFocus(selectedComponentElement != null);
        }
        this.selectedAreaElement = area;
    }

    private void toggleAreaVisibility(MgnlArea area, boolean visible) {
        MgnlArea parentArea = area.getParentArea();
        if (parentArea != null) {
            toggleAreaVisibility(parentArea, visible);
            toggleChildComponentVisibility(parentArea, visible);

        }
        // root areas are always visible
        if (!model.getRootAreas().contains(area)) {
            area.setVisible(visible);
        }

        toggleNestedAreasVisibility(area, visible);
    }

    private void toggleChildComponentVisibility(MgnlArea area, boolean visible) {
        // do not hide empty root areas placeholder
        if (!model.getRootAreas().contains(area) || !area.getComponents().isEmpty()) {
            area.setPlaceHolderVisible(visible);
        }

        // hide
        if (!visible && !area.getComponents().isEmpty()) {
            area.setPlaceHolderVisible(false);
        }

        for (MgnlComponent component : area.getComponents()) {
            // toggle all child-components editbar visibility - does this case occur?
            component.setVisible(visible);

            // toggle all child-components-area visibility
            for (MgnlArea childArea : component.getAreas()) {
                childArea.setVisible(visible);
                toggleNestedAreasVisibility(childArea, visible);
            }
        }
    }

    private void toggleNestedAreasVisibility(MgnlArea area, boolean visible) {
        for (MgnlArea childArea : area.getAreas()) {
            childArea.setVisible(visible);
            toggleNestedAreasVisibility(childArea, visible);
        }
    }

    /**
     * This toggles the 'initial' focus on the pages root-area elements. That's the light-greenish coloring when the page
     * is loaded or the page gets selected.
     */
    private void togglePageFocus(boolean visible) {
        for (MgnlArea root : model.getRootAreas()) {
            root.toggleInitFocus(visible);
        }
    }

    private void dispatchElementSelection(MgnlElement mgnlElement) {
        mgnlElement = (mgnlElement != null) ? mgnlElement : model.getRootPage();
        if (mgnlElement != null) {
            eventBus.fireEvent(new SelectElementEvent(mgnlElement.getTypedElement()));
        }
    }
}
