/**
 * This file Copyright (c) 2011-2018 Magnolia International
 * Ltd.  (http://www.magnolia-cms.com). All rights reserved.
 *
 *
 * This file is dual-licensed under both the Magnolia
 * Network Agreement and the GNU General Public License.
 * You may elect to use one or the other of these licenses.
 *
 * This file is distributed in the hope that it will be
 * useful, but AS-IS and WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE, TITLE, or NONINFRINGEMENT.
 * Redistribution, except as permitted by whichever of the GPL
 * or MNA you select, is prohibited.
 *
 * 1. For the GPL license (GPL), you can redistribute and/or
 * modify this file under the terms of the GNU General
 * Public License, Version 3, as published by the Free Software
 * Foundation.  You should have received a copy of the GNU
 * General Public License, Version 3 along with this program;
 * if not, write to the Free Software Foundation, Inc., 51
 * Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * 2. For the Magnolia Network Agreement (MNA), this file
 * and the accompanying materials are made available under the
 * terms of the MNA which accompanies this distribution, and
 * is available at http://www.magnolia-cms.com/mna.html
 *
 * Any modifications to this file must keep this entire header
 * intact.
 *
 */
package info.magnolia.ui.vaadin.gwt.client.widget.controlbar;

import info.magnolia.jcr.util.NodeTypes;

import com.google.gwt.dom.client.Element;
import com.google.gwt.user.client.DOM;
import com.google.gwt.user.client.ui.FlowPanel;
import com.google.gwt.user.client.ui.Widget;

/**
 * Base class for horizontal bars with buttons.
 * The instantiation of control-bars needs to happen in the following order:
 * <pre>
 *     <ul>
 *         <li>
 *             {@link #initLayout()} sets up the layout and the widgets in the necessary order: The buttons are placed
 *             first in the DOM, followed by the status-indicator and the label.
 *         </li>
 *         <li>
 *             {@link #addLabel(String, int)} or {@link #addStatusIndicator(int)} as these are depending on their div being
 *             present in the DOM, when added. If not called the divs will remain empty and without any class-name.
 *         </li>
 *     </ul>
 * </pre>
 */
public abstract class AbstractBar extends SimpleBar {


    private final static String EDITOR_BAR_LABEL_CLASS_NAME = "mgnlEditorBarLabel";
    private final static String EDITOR_BAR_BUTTONS_CLASS_NAME = "mgnlEditorBarButtons";
    private final static String EDITOR_BAR_WITH_STATUS_CLASS_NAME = "mgnlEditorBarStatusIndicator";

    private final static String MGNL_LEVEL_CLASS_NAME = "mgnlLevel-";

    protected final static String AREA_CLASS_NAME = "area";
    protected final static String COMPONENT_CLASS_NAME = "component";

    protected final static String ICON_CLASS_NAME = "editorIcon";
    protected final static String EDIT_CLASS_NAME = "icon-edit";
    protected final static String ADD_CLASS_NAME = "icon-add-item";

    private final static String STATUS_INDICATOR_CLASS_NAME = "status-indicator";
    private final static String STATUS_INACTIVATED_CLASS_NAME = "background-color-red icon-status-red";
    private final static String STATUS_MODIFIED_CLASS_NAME = "background-color-yellow icon-status-orange";

    private final static int MAX_LEVEL = 6;

    private FlowPanel buttons;
    private Element label;
    private Element activationIndicator;

    /**
     * Initializes the layout. The order of the elements matter.
     * IIRC the FlowPanel is needed to propagate click- and touch-events, otherwise a div vould be used for the buttons
     * as well.
     */
    protected void initLayout() {
        this.buttons = new FlowPanel();
        this.activationIndicator = DOM.createDiv();
        this.label = DOM.createDiv();

        buttons.setStylePrimaryName(EDITOR_BAR_BUTTONS_CLASS_NAME);
        insert(buttons, 0); // must be first element to allow overflowing (ellipsis) the label

        getElement().appendChild(activationIndicator);
        getElement().appendChild(label);
    }

    /**
     * Adds the label to the control-bar. Sets the css class-names and adds the level-padding.
     */
    protected void addLabel(String text, int level) {
        label.setClassName(EDITOR_BAR_LABEL_CLASS_NAME);
        label.setInnerText(text);
        // tooltip. Nice to have when area label is truncated because too long.
        label.setTitle(text);
        // replace the numerical level with 'max' if the #MAX_LEVEL has been reached
        label.addClassName(MGNL_LEVEL_CLASS_NAME + (level > MAX_LEVEL ? "max" : String.valueOf(level)));
    }

    /**
     * Adds the status-indicator to the control-bar and the correlating css classes-names to the status-div as well
     * as the control-bar itself for the left indentation.
     */
    protected void addStatusIndicator(int activationStatus) {
        switch (activationStatus) {
        case NodeTypes.Activatable.ACTIVATION_STATUS_MODIFIED:
            activationIndicator.setClassName(STATUS_INDICATOR_CLASS_NAME);
            activationIndicator.addClassName(STATUS_MODIFIED_CLASS_NAME);
            addStyleName(EDITOR_BAR_WITH_STATUS_CLASS_NAME);
            break;
        case NodeTypes.Activatable.ACTIVATION_STATUS_NOT_ACTIVATED:
            activationIndicator.setClassName(STATUS_INDICATOR_CLASS_NAME);
            activationIndicator.addClassName(STATUS_INACTIVATED_CLASS_NAME);
            addStyleName(EDITOR_BAR_WITH_STATUS_CLASS_NAME);
            break;
        // otherwise nothing, this is only run once, so no need to remove anything.
        case NodeTypes.Activatable.ACTIVATION_STATUS_ACTIVATED:
        default:
            break;
        }
    }

    @Override
    public void onAttach() {
        super.onAttach();
    }

    protected void addButton(final Widget button) {
        buttons.add(button);
    }
}
